from accrete.forms import TenantModelForm
from .models import Sequence


class SequenceCreateForm(TenantModelForm):
    class Meta:
        model = Sequence
        fields = [
            'name',
            'nextval',
            'step'
        ]



import logging
from uuid import uuid4
from functools import partial

from django import forms
from django.contrib.auth import get_user_model
from django.db import transaction
from accrete.tenant import get_tenant

_logger = logging.getLogger(__name__)
User = get_user_model()


class One2ManyModelForm(forms.ModelForm):
    """
    Template Access:
    In Templates/Forms the generated fields can be accessed via
    form.get_>>related_name<< e.g. form.get_items returns
    a list of dictionaries containing the fields.

    Cleaning:
    For every One2Many field, clean_>>related_name<< is called.
    To clean all entries together, self.cleaned_data[>>related_name<<] is
    available in clean() containing a list of all entries

    Initial data can be supplied by creating a form method called
    get_initial_>>related_name<<_>>field_name<< which takes the
    related instance as a parameter
    or by setting an attribute on the model which can be a callable


    One2Many field configuration:
    o2m_fields = {
        # related_name
        'items': {
            # field that holds the key to self.model
            'fk_field': 'pricelist',
            # Order by the given fields value, can be prefixed by '-' to
            # control ascending/descending order
            'order': '-pk'
            # fields of the related model that are needed in the form
            'fields': {
                # pk field must be present for updating existing rows.
                # Otherwise, all related objects will be deleted and recreated
                # upon saving.
                'pk': {
                    'field_class': partial(
                        forms.IntegerField,
                        required=False,
                        widget=forms.HiddenInput
                    ),
                    # if only data is sent with fields that have
                    # "count_as_empty" set to true, the whole fieldset is
                    # ignored and won't be added to self.fields
                    'count_as_empty': True
                },
                'product': {
                    'field_class': partial(
                        forms.ModelChoiceField,
                        Product.objects.all(),
                        label=_('Product')
                    )
                },
                'price': {
                    'field_class': partial(
                        forms.FloatField,
                        label=_('Price')
                    )
                }
            }
        }
    }
    """
    o2m_fields = {}

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.o2m_keys = None
        if self.instance and self.instance.pk and not kwargs.get('data'):
            self.add_instance_o2m_objects()
        else:
            self.add_o2m_fields(**kwargs)
        self.add_o2m_getter()

    def add_instance_o2m_objects(self):
        self.o2m_keys = set()
        for field in self.o2m_fields.keys():
            related_objects = getattr(self.instance, field)
            for o in related_objects.all():
                field_key = f'{field}_{str(uuid4())[:8]}'
                for k, v in self.o2m_fields[field]['fields'].items():
                    func_name = f'get_initial_{field}_{k}'
                    field_name = f'{field_key}-{k}'
                    self.fields[field_name] = v['field_class']()
                    if hasattr(self, func_name):
                        initial = getattr(self, func_name)(o)
                    else:
                        initial = getattr(o, k) if hasattr(o, k) else False
                    if callable(self.fields[field_name].initial):
                        initial = initial()
                    self.fields[field_name].initial = initial
                self.o2m_keys.add(field_key)

    def add_o2m_fields(self, **kwargs):
        self.o2m_keys = set()
        data = kwargs.get('data', {})
        for field, config in self.o2m_fields.items():
            fields = config['fields']
            field_keys = set(
                k.split('-')[0] for k in data.keys() if k.startswith(field)
            )
            if not field_keys:
                field_keys = set(
                    f'{k}_1' for k in self.o2m_fields.keys()
                )
            for field_key in field_keys:
                to_add = any([
                    data.get(f'{field_key}-{name}')
                    for name in list(fields.keys())
                    if not fields.get(f'{name}', {}).get('count_as_empty')
                ])
                if to_add:
                    for k, v in fields.items():
                        fname = f'{field_key}-{k}'
                        self.fields[fname] = v['field_class']()
                        self.fields[fname].initial = kwargs.get(fname)
                    self.o2m_keys.add(field_key)

    def add_o2m_getter(self):
        for k in self.o2m_fields.keys():
            func_name = f'get_{k}'
            setattr(self, func_name, partial(self.get_o2m_fields, k))

    def get_o2m_fields(self, prefix):
        fields = []
        for k in self.o2m_keys:
            if k.startswith(prefix):
                fnames = list(
                    self.o2m_fields.get(prefix, {}).get('fields', [])
                )
                fields.append({
                    fname: self[f'{k}-{fname}']
                    for fname in fnames
                })

        order = self.o2m_fields.get(prefix, {}).get('order')
        reverse = False
        if order and order.startswith('-'):
            order = order[1:]
            reverse = True
        if order and fields:
            fields = sorted(
                fields, key=lambda x: x[order].value(), reverse=reverse
            )
        return fields

    def _clean_fields(self):
        super()._clean_fields()
        self._clean_o2m_fields()

    def _clean_o2m_fields(self):
        for related_name, related_field_config in self.o2m_fields.items():
            self.cleaned_data[related_name] = []
            fields = related_field_config['fields'].keys()
            entries = filter(
                lambda x: x.startswith(related_name), self.o2m_keys
            )
            for entry in entries:
                value = {
                    fname: self.cleaned_data.get(f'{entry}-{fname}')
                    for fname in fields
                }
                try:
                    self.cleaned_data[related_name].append(
                        getattr(self, f'clean_{related_name}')(value)
                    )
                except AttributeError:
                    self.cleaned_data[related_name].append(value)
                except forms.ValidationError as error:
                    self.add_error(entry, error)

    def save_o2m(self):
        for field in self.o2m_fields.keys():
            fk_field = self.o2m_fields[field].get('fk_field')
            if not fk_field:
                raise KeyError(
                    'One2Many Form definition is missing the '
                    f'"fk_field" attribute on field {field}'
                )

            related_objects = getattr(self.instance, field)
            if not self.cleaned_data.get(field, []):
                related_objects.all().delete()

            related_object_fields = [
                field.name for field
                in related_objects.model._meta.get_fields()
            ]
            update_pks, to_update, to_create = [], [], []

            for item in self.cleaned_data.get(field, []):
                item.update({fk_field: self.instance})
                if pk := item.get('pk'):
                    update_pks.append(pk)
                    to_update.append(item)
                else:
                    to_create.append(item)

            qs = related_objects.filter(pk__in=update_pks).order_by('pk')
            related_objects.exclude(pk__in=update_pks).delete()
            to_update = sorted(to_update, key=lambda x: x['pk'])
            if qs.count() != len(to_update):
                raise ValueError(
                    'Mismatching length of found objects and update data'
                )
            updated = []
            for obj, update in zip(qs, to_update):
                for k, v in update.items():
                    if k not in related_object_fields:
                        continue
                    setattr(obj, k, v)
                    updated.append(obj)
            objects_to_create = []
            for c in to_create:
                values = {}
                for k, v in c.items():
                    if k in related_object_fields:
                        values.update({k: v})
                objects_to_create.append(related_objects.model(**values))
            if updated:
                field_list = [
                    f for f in to_update[0].keys()
                    if f in related_object_fields
                ]
                related_objects.bulk_update(
                    updated, fields=field_list
                )
            if to_create:
                related_objects.bulk_create(objects_to_create)

    def save(self, commit=True):
        super().save(commit=commit)
        if commit:
            self.save_o2m()
        return self.instance


class TenantForm(forms.Form):

    def __init__(self, *args, **kwargs):
        self.tenant = get_tenant()
        super().__init__(*args, **kwargs)
        fields_to_filter = filter(
            lambda x:
            hasattr(x, 'queryset') and hasattr(x.queryset.model, 'tenant'),
            self.fields.values()
        )
        for field in fields_to_filter:
            field.queryset = field.queryset.filter(tenant=self.tenant)


class TenantModelForm(One2ManyModelForm):

    def __init__(self, *args, **kwargs):
        self.tenant = get_tenant()
        super().__init__(*args, **kwargs)
        fields_to_filter = filter(
            lambda x:
            hasattr(x, 'queryset') and hasattr(x.queryset.model, 'tenant'),
            self.fields.values()
        )
        for field in fields_to_filter:
            field.queryset = field.queryset.filter(tenant=self.tenant)

    def save_o2m(self):
        o2m_fields = self.o2m_fields.keys()
        for field in o2m_fields:
            for item in self.cleaned_data.get(field, []):
                item.update({'tenant': self.tenant})
        return super().save_o2m()

    def save(self, commit=True):
        super().save(commit=False)
        if not self.instance.pk or not self.instance.tenant:
            self.instance.tenant = self.tenant

        if commit:
            self.instance.save()
            self.save_m2m()
            self.save_o2m()

        return self.instance


def save_form(form, reraise=False):
    form.is_saved = False
    form.save_error = None
    form.save_error_id = None
    try:
        if form.is_valid():
            with transaction.atomic():
                form.save()
            form.is_saved = True
    except Exception as e:
        form.save_error = repr(e)
        error_id = str(uuid4())[:8]
        _logger.exception(f'{error_id}: {e}')
        form.save_error_id = error_id
        if reraise:
            raise e
    return form



import logging
from accrete import models
from accrete.tenant import get_tenant

_logger = logging.getLogger(__name__)


def is_member(tenant, user):
    return tenant.members.filter(user=user, is_active=True).exists()


def members_for_current_tenant():
    tenant = get_tenant()
    return tenant and tenant.members or models.Member.objects.none()


def all_tenants():
    return models.Tenant.objects.all()



import logging
from dataclasses import dataclass, field

from django.utils.translation import gettext_lazy as _
from django.db.models import Model, QuerySet, Q
from django.core.paginator import Paginator
from django.forms import Form, ModelForm
from accrete.querystring import parse_querystring

from .querystring import load_querystring, build_querystring
from .components import ClientAction, BreadCrumb, TableField
from .filter import Filter

_logger = logging.getLogger(__name__)

DEFAULT_PAGINATE_BY = 40


@dataclass
class ListContext:

    model: type[Model]
    get_params: dict
    title: str = None
    context: dict = field(default_factory=dict)
    queryset: QuerySet = None
    select_related: list[str] = field(default_factory=list)
    prefetch_related: list[str] = field(default_factory=list)
    annotate: dict = field(default_factory=dict)
    paginate_by: int = DEFAULT_PAGINATE_BY
    order_by: list[str] = field(default_factory=list)
    filter_relation_depth: int = 4
    default_filter_term: str = ''
    actions: list[ClientAction] = field(default_factory=list)
    breadcrumbs: list[BreadCrumb] = field(default_factory=list)
    object_label: str = None
    fields: list[TableField] = field(default_factory=list)
    unselect_button: bool = False
    endless_scroll: bool = True

    def get_queryset(self):
        if self.queryset:
            return self.queryset

        order = self.order_by or self.model._meta.ordering

        # pks = self.model.objects.filter(
        #     parse_querystring(self.model, self.get_params.get('q', '[]'))
        # ).distinct().values_list('pk', flat=True)

        queryset = self.model.objects.select_related(
            *self.select_related
        ).prefetch_related(
            *self.prefetch_related
        ).filter(
            parse_querystring(self.model, self.get_params.get('q', '[]'))
        ).annotate(
            **self.get_annotations()
        ).order_by(
            *order
        ).distinct()

        return queryset

    def get_annotations(self):
        annotations = {
            annotation['name']: annotation['func']
            for annotation in getattr(self.model, 'annotations', [])
        }
        if self.annotate:
            annotations.update(self.annotate)
        return annotations

    def get_page_number(self, paginator):
        page_number = self.get_params.get('page', '1')

        try:
            page_number = int(page_number)
        except ValueError:
            page_number = 1

        if page_number < 1:
            page_number = 1
        elif page_number > paginator.num_pages:
            page_number = paginator.num_pages
        return page_number

    def get_paginate_by(self):
        paginate_by = self.get_params.get('paginate_by', self.paginate_by)
        try:
            paginate_by = int(paginate_by)
        except ValueError:
            paginate_by = self.paginate_by
        return paginate_by

    def dict(self):
        queryset = self.get_queryset()
        paginate_by = self.get_paginate_by()
        paginator = Paginator(queryset, paginate_by)
        page = paginator.page(self.get_page_number(paginator))
        context = {
            'queryset': queryset,
            'paginate_by': paginate_by,
            'order_by': self.get_params.get('order_by', self.model._meta.ordering),
            'paginator': paginator,
            'page': page,
            'list_pagination': True,
            'title': self.title or self.model._meta.verbose_name_plural,
            'object_label': self.object_label or self.model._meta.verbose_name or _('Name'),
            'filter': Filter(self.model, self.filter_relation_depth),
            'default_filter_term': self.default_filter_term,
            'breadcrumbs': self.breadcrumbs,
            'querystring': load_querystring(self.get_params),
            'url_params': build_querystring(self.get_params),
            'actions': self.actions,
            'fields': self.fields,
            'endless_scroll': self.endless_scroll
        }
        context.update(self.context)
        return context


@dataclass
class DetailContext:

    obj: Model | type[Model]
    get_params: dict
    order_by: str = None
    paginate_by: int = DEFAULT_PAGINATE_BY
    title: str = None
    queryset: type[QuerySet] = None
    select_related: list[str] = field(default_factory=list)
    prefetch_related: list[str] = field(default_factory=list)
    annotate: dict = field(default_factory=dict)
    actions: list[ClientAction] = field(default_factory=list)
    breadcrumbs: list[BreadCrumb] = field(default_factory=list)
    context: dict = field(default_factory=dict)

    def get_queryset(self):
        if self.queryset:
            return self.queryset

        order = self.order_by or self.obj._meta.model._meta.ordering

        pks = self.obj._meta.model.objects.filter(
            parse_querystring(self.obj._meta.model, self.get_params.get('q', '[]'))
        ).distinct().values_list('pk', flat=True)

        queryset = self.obj._meta.model.objects.select_related(
            *self.select_related
        ).prefetch_related(
            *self.prefetch_related
        ).filter(
            Q(pk__in=pks)
        ).annotate(
            **self.get_annotations()
        ).order_by(
            *order
        ).distinct()

        return queryset

    def get_annotations(self):
        annotations = {
            annotation['name']: annotation['func']
            for annotation in getattr(self.obj._meta.model, 'annotations', [])
        }
        if self.annotate:
            annotations.update(self.annotate)
        return annotations

    def get_paginate_by(self):
        paginate_by = self.get_params.get('paginate_by', self.paginate_by)
        try:
            paginate_by = int(paginate_by)
        except ValueError:
            paginate_by = self.paginate_by
        return paginate_by

    def get_pagination_context(self):
        if not hasattr(self.obj, 'get_absolute_url'):
            _logger.warning(
                'Detail pagination disabled for models without the '
                'get_absolute_url attribute. Set paginate_by to 0 to '
                'deactivate pagination.'
            )
            return {}
        queryset = self.get_queryset()
        idx = (*queryset,).index(self.obj)
        previous_object_url = (
            queryset[idx - 1] if idx - 1 >= 0 else queryset.last()
        ).get_absolute_url()
        next_object_url = (
            queryset[idx + 1] if idx + 1 <= queryset.count() - 1 else queryset.first()
        ).get_absolute_url()
        ctx = {
            'previous_object_url': previous_object_url,
            'next_object_url': next_object_url,
            'current_object_idx': idx + 1,
            'total_objects': queryset.count(),
            'detail_pagination': True
        }
        return ctx

    def dict(self):
        paginate_by = self.get_paginate_by()
        ctx = {
            'object': self.get_queryset().get(pk=self.obj.pk),
            'title': self.title or self.obj,
            'order_by': self.get_params.get('order_by', self.obj._meta.model._meta.ordering),
            'paginate_by': paginate_by,
            'detail_pagination': False,
            'breadcrumbs': self.breadcrumbs,
            'url_params': build_querystring(self.get_params, ['page']),
            'actions': self.actions
        }
        if self.paginate_by > 0:
            ctx.update(self.get_pagination_context())
        ctx.update(self.context)
        return ctx


@dataclass
class FormContext:

    model: Model | type[Model]
    form: Form | ModelForm
    get_params: dict
    title: str = None
    context: dict = field(default_factory=dict)
    form_id: str = 'form'
    add_default_actions: bool = True
    discard_url: str = None
    actions: list[ClientAction] = field(default_factory=list)
    breadcrumbs: list[BreadCrumb] = field(default_factory=list)

    def get_default_form_actions(self):
        actions = [
            ClientAction(
                name=_('Save'),
                submit=True,
                class_list=['is-success'],
                form_id=self.form_id
            )
        ]
        try:
            url = self.discard_url or (self.model.pk and self.model.get_absolute_url())
        except TypeError:
            raise TypeError(
                'Supply the discard_url parameter if FormContext is called '
                'with a model class instead of an instance.'
            )
        except AttributeError as e:
            _logger.error(
                'Supply the discard_url parameter if FormContext is '
                'called with a model instance that has the get_absolute_url '
                'method not defined.'
            )
            raise e

        actions.append(
            ClientAction(
                name=_('Discard'),
                url=url,
            )
        )
        return actions

    def get_title(self):
        if self.title:
            return self.title
        try:
            int(self.model.pk)
            return _('Edit')
        except TypeError:
            return _('Add')

    def dict(self):
        ctx = {
            'title': self.get_title(),
            'form': self.form,
            'form_id': self.form_id,
            'url_params': build_querystring(self.get_params, ['page']),
            'actions': [],
            'breadcrumbs': self.breadcrumbs,
        }
        if self.add_default_actions:
            ctx.update({'actions': self.get_default_form_actions()})
        ctx['actions'].extend(self.actions)
        ctx.update(self.context)
        return ctx



from django.conf import settings
from django.core.validators import validate_email
from django.utils.translation import gettext_lazy as _
from django.forms import ModelForm, fields, ValidationError, PasswordInput, EmailInput
from .models import User


class UserForm(ModelForm):

    language_code = fields.ChoiceField(
        label=_('Language'),
        choices=settings.LANGUAGES,
        required=True
    )

    class Meta:
        model = User
        fields = [
            'username',
            'first_name',
            'last_name',
        ]

    def save(self, commit=True):
        super().save(commit=False)
        self.instance.language_code = self.cleaned_data['language_code']
        if commit:
            self.instance.save()
            self.save_m2m()


class ChangePasswordForm(ModelForm):

    old_password = fields.CharField(
        label=_('Old password'),
        widget=PasswordInput(render_value=True),
        max_length=128,
    )

    new_password = fields.CharField(
        label=_('New password'),
        widget=PasswordInput(render_value=True),
        max_length=128
    )

    new_password_confirm = fields.CharField(
        label=_('New password confirmation'),
        widget=PasswordInput(render_value=True),
        max_length=128
    )

    class Meta:
        model = User
        fields = []

    def clean_old_password(self):
        old_password = self.cleaned_data['old_password']
        if not self.instance.check_password(old_password):
            raise ValidationError(_(
                'Password Validation failed'
            ))
        return old_password

    def clean(self):
        super().clean()
        new_password = self.cleaned_data.get('new_password')
        new_password_confirm = self.cleaned_data.get('new_password_confirm')
        if not new_password or not new_password_confirm:
            raise ValidationError(_(
                'New password and password confirmation are required'
            ))
        if new_password != new_password_confirm:
            raise ValidationError(
                _('New password and confirmation did not match.')
            )
        return self.cleaned_data

    def save(self, commit=True):
        user = self.instance
        user.set_password(self.cleaned_data['new_password'])
        user.save()


class ChangeEmailForm(ModelForm):

    email = fields.EmailField(
        label=_('New Email Address'),
        widget=EmailInput(),
    )

    password = fields.CharField(
        label=_('Password'),
        widget=PasswordInput()
    )

    class Meta:
        model = User
        fields = []

    def clean_email(self):
        email = self.cleaned_data.get('email', '').lower()
        validate_email(email)
        if User.objects.filter(email=email):
            raise ValidationError(
                _('This email address is already bound to an account.')
            )
        return email

    def clean_password(self):
        password = self.cleaned_data.get('password')
        if not self.instance.check_password(password):
            raise ValidationError(_(
                'Password incorrect'
            ))
        return password

    def save(self, commit=True):
        self.instance.email = self.cleaned_data['email']
        self.instance.save()



import logging
import json
import operator
from django.db.models import Model, Q, QuerySet

_logger = logging.getLogger(__name__)


def filter_querystring(model: type[Model], query_string: str) -> QuerySet:
    return model.objects.filter(parse_querystring(model, query_string))


def parse_querystring(model: type[Model], query_string: str) -> Q:
    """
    param: query_string: JSON serializable string
    [{term: value}, '&', [{term: value}, '|', {~term: value}]]
    Q(term=value) & (Q(term=value) | ~Q(term=value))
    """

    def get_expression(term: str, value) -> Q:
        invert = False
        if term.startswith('~'):
            invert = True
            term = term[1:]

        parts = term.split('_a_')
        if len(parts) == 1:
            expression = Q(**{term: value})
            return ~expression if invert else expression

        rel_path = parts[0].rstrip('__')
        term = parts[1]
        rel_model = get_related_model(rel_path) if rel_path else model
        objects = rel_model.objects.annotate(**{
            annotation['name']: annotation['func']
            for annotation in rel_model.annotations
        }).filter(Q(**{term: value}))
        expression = Q(**{
            f'{rel_path}{"__" if rel_path else ""}id__in': objects.values_list('id', flat=True)
        })

        return ~expression if invert else expression

    def get_related_model(rel_path: str):
        related_model = model
        for part in rel_path.split('__'):
            try:
                related_model = related_model._meta.fields_map[part].related_model
            except (AttributeError, KeyError):
                try:
                    related_model = getattr(related_model, part).field.related_model
                except AttributeError:
                    break
        return related_model

    def parse_query_block(sub_item) -> Q:
        op = ops['&']
        parsed_query = Q()
        for item in sub_item:
            if isinstance(item, list):
                parsed_query = op(parsed_query, parse_query_block(item))
                op = ops['&']
            elif isinstance(item, dict):
                dict_query = Q()
                for term, value in item.items():
                    dict_query = ops['&'](dict_query, get_expression(term, value))
                parsed_query = op(parsed_query, dict_query)
            elif isinstance(item, str):
                if item not in '&|^':
                    raise ValueError(
                        f'Invalid operator in querystring: {item}.'
                        f'Operator must be one of &, |, ^'
                    )
                op = ops[item]

            else:
                raise ValueError(
                    f'Unsupported item in querystring: {item}.'
                    f'Item must be an instance of list, dict or str'
                )
        return parsed_query

    query_data = json.loads(query_string)
    if isinstance(query_data, dict):
        query_data = [query_data]

    ops = {'&': operator.and_, '|': operator.or_, '^': operator.xor}
    query = parse_query_block(query_data)
    return query



import contextvars
import logging
import time
from django.apps import apps
from django.db.models import Q

_logger = logging.getLogger(__name__)

tenant_contextvar = contextvars.ContextVar('tenant', default=None)
member_contextvar = contextvars.ContextVar('member', default=None)


def set_tenant(tenant):
    tenant_contextvar.set(tenant)


def get_tenant():
    return tenant_contextvar.get()


def set_member(member):
    tenant = member.tenant if member is not None else None
    member_contextvar.set(member)
    tenant_contextvar.set(tenant)


def get_member():
    return member_contextvar.get()


class Unscoped:

    def __init__(self, tenant):
        self.tenant = tenant

    def __enter__(self):
        if self.tenant is None:
            _logger.warning(
                'Entering unscoped context manager with tenant already set to None!'
            )
        set_tenant(None)

    def __exit__(self, exc_type, exc_val, exc_tb):
        set_tenant(self.tenant)


def unscoped():
    return Unscoped(get_tenant())


def per_tenant(include: Q = None, exclude: Q = None):
    def decorator(f):
        def wrapper(*args, **kwargs):
            tenants = apps.get_model('accrete', 'Tenant').objects.all()
            if include is not None:
                tenants = tenants.filter(include)
            if exclude is not None:
                tenants = tenants.exclude(exclude)
            _logger.info(f'Running {f.__module__}.{f.__name__}')
            start_time = time.time()
            for tenant in tenants:
                set_tenant(tenant)
                f(*args, **kwargs)
            _logger.info(
                f'Finished {f.__module__}.{f.__name__} in {time.time() - start_time} seconds.'
            )
        return wrapper
    return decorator

